
from vivarium.library.schema import array_from, array_to
import numpy as np

def one_to_one_map(bsp1, bsp2, species_map):
    #Takes two Bioscrape Processes, bsp1 and bsp2 and
    # species_map: dictionar(a species in bsp1: a species in bsp2)
    #Produces the mapping species1[i] --> species2[i]

    all_species1 = bsp1.get_model_species_ids()
    all_species2 = bsp2.get_model_species_ids()

    projection = np.zeros((len(all_species1), len(all_species2)))

    #Create the projection matrix
    for s1 in species_map:
        if s1 not in all_species1:
            raise ValueError(f"{s1} not found in Bioscrape Process {bsp1.sbml_file}.")
        else:
            ind1 = all_species1.index(s1)

        s2 = species_map[s1]
        if s2 not in all_species2:
            raise ValueError(f"{s2} not found in Bioscrape Process {bsp2.sbml_file}.")
        else:
            ind2 = all_species2.index(s2)

        projection[ind1, ind2] = 1

    def map_function(states):
        source_delta_array = array_from(states['source_deltas'])
        output_delta_array = np.dot(source_delta_array, projection)
        return array_to(all_species2, output_delta_array)

    return map_function


def one_to_many_map(bsp1, bsp2, species_map, proportional = True):
    #Takes two Bioscrape Processes, bsp1 and bsp2 and
    #species_map { species in bsp1: [list of species in bsp2] }
    #Produces the mapping species1[i] --> [species2[i]]
    #proprtional: True/False
    #    if True the amount of species1[i] mapped to species2[i][j] is proportianal to the amount of species2[i][j]
    #        one_to_many_map(species1[i]) ~ species2[i][j]/sum(species[2][i])
    #    if False: the amount of species1[i] mapped to species2[i][j] ~ 1/len(species[2][i])

    all_species1 = bsp1.get_model_species_ids()
    all_species2 = bsp2.get_model_species_ids()

    projection = np.zeros((len(all_species1), len(all_species2)))

    #Create the projection matrix
    for s1 in species_map:
        if s1 not in all_species1:
            raise ValueError(f"{s1} not found in Bioscrape Process {bsp1.sbml_file}.")
        else:
            ind1 = all_species1.index(s1)

        for s2 in species_map[s1]:
            if s2 not in all_species2:
                raise ValueError(f"{s2} not found in Bioscrape Process {bsp2.sbml_file}.")
            else:
                ind2 = all_species2.index(s2)

            projection[ind1, ind2] = 1


    def map_function(states):
        source_delta_array = array_from(states['source_deltas'])

        if not proportional:
            normalizer = 1/np.sum(projection, 1)
            normalized_proj = normalizer[:, np.newaxis]*projection
        else:
            #Normalize the projection matrix based upon the total concentrations of each target species
            output_array = array_from(states['target_state'])
            normalized_proj = projection*output_array
            normalizer = np.sum(normalized_proj, 1)[:, np.newaxis]*projection
            normalized_proj = normalized_proj/(normalizer+(normalizer == 0))

            #Cycle through projections, find degenerate cases
            #Degenerate cases occur when none of the many species are present
            #In this case, the normalizer is uniform.
            for i in range(projection.shape[0]):
                if np.sum(normalizer[i, :]) == 0 and np.sum(projection[i, :]) > 0:
                    normalizer[i, :] = sum(projection[i, :])
                    normalized_proj[i, :] = projection[i, :]/normalizer[i, :]

        output_delta_array = np.dot(source_delta_array, normalized_proj)
        return array_to(all_species2, output_delta_array)

    return map_function


def many_to_one_map(bsp1, bsp2, species_map):
    #Takes two Bioscrape Processes, bsp1 and bsp2 and
    #species_map = { species in bsp2 : [list of many species in bsp1] }
    #Produces the mapping sum(species1[i]) --> species2[i]

    all_species1 = bsp1.get_model_species_ids()
    all_species2 = bsp2.get_model_species_ids()

    projection = np.zeros((len(all_species1), len(all_species2)))

    #Create the projection matrix
    for s2 in species_map:

        if s2 not in all_species2:
            raise ValueError(f"{s2} not found in Bioscrape Process {bsp2.sbml_file}.")
        else:
            ind2 = all_species2.index(s2)

        for s1 in species_map[s2]:
            if s1 not in all_species1:
                raise ValueError(f"{s1} not found in Bioscrape Process {bsp1.sbml_file}.")
            else:
                ind1 = all_species1.index(s1)

            projection[ind1, ind2] = 1

    def map_function(states):
        source_delta_array = array_from(states['source_deltas'])
        output_delta_array = np.dot(source_delta_array, projection)
        return array_to(all_species2, output_delta_array)

    return map_function


def stochiometric_map(parent_func, stochiometric_dictionary):
    #This converts any map function into a map that obeys stochiometric laws
    # parent_func is any parent mapping
    # stochiometric_dictionary is a dictionary: {species : {coupled species : stochiometric coefficient} }
    # Any species may have one or more coupled species (which always change along with it) 
    # using arbitrary stochiometric coefficients
    def map_function(states):
        deltas_dict = parent_func(states)
        new_deltas = dict(deltas_dict)
        for s in deltas_dict:
            if s in stochiometric_dictionary:
                for s2 in stochiometric_dictionary[s]:
                    coef = stochiometric_dictionary[s][s2]
                    if s2 in deltas_dict:
                        new_deltas[s2] += deltas_dict[s]*coef
                    else:
                        new_deltas[s2] = deltas_dict[s]*coef

        return new_deltas

    return map_function

def det_to_stoch_map(parent_func):
    #This converts any parent_func to produce integer changes in the species counts
    #using the volume of the underlying bioscrape processes. This allows for deterministic
    #and stochastic processes to be run together. Note: rounding errors may occur.

    def map_function(states):
        deltas_dict = parent_func(states)
        if 'global' in states and 'source_volume' in states['global']:
            V_source = states['global']['source_volume']
        else:
            V_source = 1.0

        for s in deltas_dict:
            sign = np.sign(deltas_dict[s])
            delta_int = np.floor(abs(deltas_dict[s]*V_source)) #Round down the change in number
            res = np.abs(deltas_dict[s]*V_source) - delta_int #Residual count

            #add/subtract an additional species with probability proportional to res.
            delta = sign*(delta_int + (np.random.random() < res))
            
            deltas_dict[s] = delta

        return deltas_dict

    return map_function

def stoch_to_det_map(parent_func):
    #This converts any parent_func that produces integer count changes into the species concentrations
    #using the volume of the underlying bioscrape processes. This allows for deterministic
    #and stochastic processes to be run together. 

    def map_function(states):
        deltas_dict = parent_func(states)
        if 'global' in states and 'target_volume' in states['global']:
            V_target = states['global']['target_volume']
        else:
            V_target = 1.0
        
        for s in deltas_dict:
            deltas_dict[s] = deltas_dict[s]/V_target #convert a change in count to a concentration

        return deltas_dict

    return map_function





'''
Execute by running: ``python -m  bioscrape.processes.bioscrape.py``

TODO: Replace the bioscrape code to implement your own process.
'''

import os
import numpy as np

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output
from bioscrape.types import Model, Volume
from bioscrape.simulator import DeterministicSimulator, ModelCSimInterface, VolumeSSASimulator, SafeModelCSimInterface
from bioscrape.lineage import LineageCSimInterface, LineageModel, LineageSSASimulator, LineageVolumeCellState

NAME = 'bioscrape'

class Bioscrape(Process):
    '''
    This process provides a wrapper around a bioscrape model, interface, and simulator.
    It allows for stochastic or determinstic simulation, variable volume, and generates ports
    to access all bioscrape species and rate parameters.
    '''

    # give the process a name, so that it can register in the process_repository
    name = NAME

    # declare default parameters as class variables
    defaults = {
        'internal_dt': 0.01,
        'stochastic': False,
        'initial_volume': 1.0,
        'safe_mode':False,
        'lineage':False
    }

    def __init__(self, parameters=None):
        super(Bioscrape, self).__init__(parameters)

        # get the parameters out of self.parameters if available, or use defaults
        if 'sbml_file' not in self.parameters and 'bioscrape_model' not in self.parameters:
            raise ValueError("Bioscrape Process requires either an sbml_file or bioscrape_model parameter.")
        elif 'sbml_file' not in self.parameters and isinstance(self.parameters['bioscrape_model'], Model):
            # load the sbml file to create the model
            self.sbml_file = None
            self.model = self.parameters['bioscrape_model']
        elif isinstance(self.parameters['sbml_file'], str) and 'bioscrape_model' not in self.parameters:
            self.sbml_file = self.parameters['sbml_file']
            if self.parameters["lineage"]:
                self.model = LineageModel(sbml_filename = self.sbml_file,)
            else:
                self.model = Model(sbml_filename = self.sbml_file, sbml_warnings = False)
        elif isinstance(self.parameters['sbml_file'], str) and isinstance(self.parameters['bioscrape_model'], Model):
            raise ValueError("Bioscrape recieved an sbml_file and a bioscrape_model. Please use one or the other.")
        else:
            raise ValueError(f"Bioscrape did not recieve a valid bioscrape_model "
                             f"(recieved: {self.parameters['bioscrape_model']} or a "
                             f"valid sbml_file (recieved: {self.parameters['sbml_file']}).")
        
        self.internal_dt = self.parameters['internal_dt']
        self.stochastic = self.parameters['stochastic']
        
        
        #Toggle using Lineage Model
        if self.parameters["lineage"]:
            if not self.stochastic:
                raise ValueError("Bioscrape lineage only available with stochastic = True")
            self.simulator = LineageSSASimulator()
            self.interface = LineageCSimInterface(self.model)
            self.volume = LineageVolumeCellState() #Create an internal bioscrape Volume
        #Otherwise use normal bioscrape models
        else: 
            # create the interface
            if self.parameters["safe_mode"]:
                self.interface = SafeModelCSimInterface(self.model, max_species_count = 10**8)
            else:
                self.interface = ModelCSimInterface(self.model)

            #Stochastic
            if self.stochastic:
                self.simulator = VolumeSSASimulator()
            #Not Stochastic
            elif not self.stochastic:
                self.interface.py_prep_deterministic_simulation()
                # create a Simulator
                self.simulator = DeterministicSimulator()
            self.volume = Volume() #Create an internal bioscrape Volume

        #Set dt
        self.interface.py_set_dt(self.internal_dt)
        #Set the volume
        self.volume.py_set_volume(self.parameters['initial_volume']) 

    def get_species_names(self):
        #Gets the names of teh species in a bioscrape model
        model_species = self.model.get_species_dictionary()
        return list(model_species.keys())

    def get_state(self, array):
        #Gets the state of a bioscrape simulation
        mapping = self.model.get_species2index()

        return {
            species: array[index]
            for species, index in mapping.items()}

    def initial_state(self, config=None):
        #gets the current (or initial) state of a bioscrape simulation.
        if config is None:
            config = {}
        self.model.set_species(config)
        state = self.model.get_species_array()
        return {
            'species': self.get_state(state)}

    def ports_schema(self):
        '''
        ports_schema returns a dictionary that declares how each state will behave.
        Each key can be assigned settings for the schema_keys declared in Store:

        * `_default`
        * `_updater`
        * `_divider`
        * `_value`
        * `_properties`
        * `_emit`
        * `_serializer`
        '''

        #Different divide settings between stochastic and determinsitic CRNs
        if self.stochastic:
            divider = "binomial"
        else:
            divider = "set" #division does not change concentrations

        return {
            'species': {
                species: {
                    '_default': 0.0,
                    '_updater': 'accumulate',
                    '_emit': True,
                    'divider':divider}
                for species in self.model.get_species()},
            'delta_species': {
                species: {
                    '_default': 0.0,
                    '_updater': 'set',
                    '_emit': False}
                for species in self.model.get_species()},
            'rates': {
                p: {
                    '_default': self.model.get_parameter_dictionary()[p],
                    '_updater':'set',
                }
                for p in self.model.get_param_list()},
            'globals': {
                'volume': {
                    '_default':self.parameters['initial_volume'],
                    '_updater': 'accumulate',
                    '_emit': True}
            }
        }

    def next_update(self, timestep, states):
        if 'species' in states:
            self.model.set_species(states['species'])
            self.interface.py_set_initial_state(self.model.get_species_array())
        if 'rates' in states:
            self.model.set_params(states['rates'])

        #Set Volume if needed
        if 'volume' in states:
            self.volume.py_set_volume(states['globals']['volume'])

        # create the interface

        timepoints = np.arange(0, timestep, self.internal_dt)
        if self.parameters["lineage"]:
            output = self.simulator.py_SimulateSingleCell(timepoints, Model = self.model, interface = self.interface, v = self.volume)
        elif self.stochastic:
            output = self.simulator.py_volume_simulate(self.interface, self.volume, timepoints)
        else:
            output = self.simulator.py_simulate(self.interface, timepoints)

        result = output.py_get_result()[-1]
        result_state = self.get_state(result)
        delta = get_delta(states['species'], result_state)
        rates = self.model.get_parameter_dictionary()

        #If the simulation is a volume simulation, return the change in volume
        if getattr(output, "py_get_volume", None) is not None:
            Vi = output.py_get_volume()[0]
            Vf = output.py_get_volume()[-1]
            deltaV = Vf-Vi
        else:
            deltaV = 0

        return {
            'species': delta,
            'delta_species': delta,
            'rates':rates,
            'globals': {'volume': deltaV}}

    def get_model(self):
        return self.model

    def get_model_species(self):
        return self.model.get_species_dictionary()

    def get_model_species_ids(self):
        return list(self.model.get_species_dictionary().keys())

    def get_volume(self):
        return self.volume.py_get_volume()


def get_delta(before, after):
    # assuming before and after have the same keys
    return {
        key: after[key] - before_value
        for key, before_value in before.items()}


def run_bioscrape_process():
    #Create a bioscrape process
    initial_parameters = {
        'sbml_file': 'Notebooks/model1.xml'}
    bioscrape_process = Bioscrape(initial_parameters)

    # run the simulation
    sim_settings = {
        'total_time': 10,
        'initial_state': bioscrape_process.initial_state()}
    output = simulate_process(bioscrape_process, sim_settings)
    

    # Return the data from the simulation.
    return output

def test_bioscrape_instantiation():

    #Deterministic Case
    bioscrape_process = Bioscrape({'sbml_file': 'Notebooks/model1.xml'})
    assert isinstance(bioscrape_process.model, Model)
    assert isinstance(bioscrape_process.interface, ModelCSimInterface)
    assert bioscrape_process.stochastic == False
    assert isinstance(bioscrape_process.simulator, DeterministicSimulator)

    #Stochastic Case
    bioscrape_process = Bioscrape({'sbml_file': 'Notebooks/model1.xml', "stochastic":True})
    assert isinstance(bioscrape_process.model, Model)
    assert isinstance(bioscrape_process.interface, ModelCSimInterface)
    assert bioscrape_process.stochastic == True
    assert isinstance(bioscrape_process.simulator, VolumeSSASimulator)

    #Custom Model Case
    M = Model(species = ["S"])
    bioscrape_process2 = Bioscrape({'bioscrape_model': M})
    assert M == bioscrape_process2.model
    assert bioscrape_process2.sbml_file is None

def test_next_update():
    initial_parameters = {
        'sbml_file': 'Notebooks/model1.xml'}
    bioscrape_process = Bioscrape(initial_parameters)

    initial_state = bioscrape_process.initial_state()
    output = bioscrape_process.next_update(1.0, initial_state)

    assert "species" in output
    assert "delta_species" in output
    assert all([output["species"][s] == output["delta_species"][s] for s in output["species"]])
    assert "rates" in output

    #set all the rates to 0
    state = {
        "rates": {p:0 for p in output["rates"]}, 
        "species":bioscrape_process.initial_state()["species"]
        }
    output2 = bioscrape_process.next_update(1.0, state)
    #nothing should change in the simulation
    assert all([output2["species"][s] == output2["delta_species"][s] ==0 for s in output2["species"]])

def test_bioscrape_process():
    '''Test that the process runs correctly.

    This will be executed by pytest.
    '''

    output = run_bioscrape_process()

    #DNA concentration should be constant
    assert all([v == output["species"]["dna_G"][0] for v in output["species"]["dna_G"]])
    #RNA concentration should be increasing
    assert all([output["species"]["rna_T"][i] < output["species"]["rna_T"][i+1] for i in range(len(output["species"]["rna_T"])-1)])
    #Protein concentration should be increasing
    assert all([output["species"]["protein_X"][i] < output["species"]["protein_X"][i+1] for i in range(len(output["species"]["protein_X"])-1)]) 

def main():
    '''Simulate the process and plot results.'''
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    output = run_bioscrape_process()

    # plot the simulation output
    plot_settings = {}
    plot_simulation_output(output, plot_settings, out_dir)


if __name__ == '__main__':
    main()




"""
===================
Bioscrape Connector
===================

run with:
> python -m vivarium_bioscrape.composites.bioscrape_connector
"""
import os

import numpy as np

# vivarium core imports
from vivarium.library.schema import array_from, array_to
from vivarium.core.experiment import pp
from vivarium.core.process import Composer
from vivarium.core.composition import (
    simulate_composer,
    COMPARTMENT_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.library.dict_utils import deep_merge

# import processes
from vivarium_bioscrape.processes.bioscrape import Bioscrape
from vivarium_bioscrape.processes.one_way_map import OneWayMap


NAME = 'bioscrape_connector'


class BioscrapeConnector(Composer):
    name = NAME
    defaults = {
        'models': {},
        'connections': [],
        }

    def __init__(self, config=None, models=None, connections=None):
        if config is None:
            config = {}
        if models is None:
            models = {}
        if connections is None:
            connections = {}

        if 'models' not in config:
            config['models'] = models
        if 'connections' not in config:
            config['connections'] = connections

        self.models = {}
        self.connections = {}
        self.connection_counts = {} #Stores the number of connections between pairs of processes, for naming purposes
        super(BioscrapeConnector, self).__init__(config)
        self.topology = self.initial_topology(self.config)

    def initial_state(self, config=None):
        """
        TODO -- automated resolution of conflicts
        """
        if config is None:
            config = {}
        self.generate()

        # set model values to config
        for name, species_dict in config.items():
            self.models[name].initial_state(species_dict)

        initial_state = {}
        for name, node in self.connections.items():
            source = node.parameters['source']
            target = node.parameters['target']
            map_function = node.map_function
            source_state = self.models[source].initial_state()
            target_state = self.models[target].initial_state()

            # Assume that source is going from 0 to the source process' initial state
            modified_source_state = {
                'source_deltas': source_state['species'],
                'target_state': target_state['species']}
            target_deltas = map_function(modified_source_state)

            for species, value in target_deltas.items():
                target_value = target_state['species'][species]
                if value != 0.0 and target_value != value:
                    raise ValueError(
                        f"initial condition ['{target}_species']['{species}']={target_value} "
                        f"does not match mapping from model {source} to model {target}")

            # set initial state according source and target species
            initial_state[f'{source}_species'] = source_state['species']
            initial_state[f'{target}_species'] = target_state['species']

        #Reset internal variables because generate_processes will be called again
        return initial_state

    def generate_processes(self, config):
        # make bioscrape processes
        
        self.connection_counts = {} #models and connections may be overwritten, so reset the counts
        for name, parameters in config['models'].items():
            self.add_model(name = name, bioscrape_parameters = parameters)

        for connection in config['connections']:
            source = connection['source']
            target = connection['target']
            map_function = connection['map_function']
            self.add_connection(source, target, map_function)

        # combine model and connection processes

        return {**self.models, **self.connections}

    def generate_topology(self, config):
        return self.topology

    def initial_topology(self, config):
        # connect the models with their stores
        models = {
            name: {
                'species': (f'{name}_species',),
                'delta_species': (f'{name}_deltas',),
                'rates': (f'{name}_rates',),
                'globals': (f'{name}_globals',)}
            for name, path in config['models'].items()}

        # make connections between model stores
        connections = {}
        for connection in config['connections']:
            source = connection['source']
            target = connection['target']

            #Count is used to allow multiple connections between models if desired
            if (source, target) in self.connection_counts:
                self.connection_counts[(source, target)] += 1
                count = self.connection_counts[(source, target)]
            else:
                count = 1
                self.connection_counts[(source, target)] = count

            connections[f'{source}_{target}_connector_{count}'] = {
                'source_deltas': (f'{source}_deltas',),
                'target_state': (f'{target}_species',),
                'globals' : {
                    f'source_volume':(f'{source}', 'globals','volume',),
                    f'target_volume':(f'{target}', 'globals','volume',),

                }
            }

        return {**models, **connections}

    def add_model(self, name = None, bioscrape_process = None, bioscrape_parameters = None):
        if name is None:
            name = str(len(self.models))

        #This test doesn't allow generate to be called multiple times
        #if name in self.models:
        #    raise ValueError(f"A model named {name} already exists!")

        if bioscrape_process is not None and bioscrape_parameters is not None:
            raise ValueError("Recieved both a bioscrape_process and bioscrape_parameters! Please use one or the other.")
        elif bioscrape_process is not None:
            if isinstance(bioscrape_process, Bioscrape):
                self.models[name] =  bioscrape_process
            else:
                raise TypeError("bioscrape_process must be a Process of type Bioscrape.")
        elif bioscrape_parameters is not None:
            if isinstance(bioscrape_parameters, dict):
                self.models[name] = Bioscrape(bioscrape_parameters)
            else:
                raise TypeError("bioscrape_parameters must be a dictionary.")
        else:
            raise ValueError("Recieved neither bioscrape_process nor bioscrape_parameters keywords. Please use one or the other (not both).")

        #Add process to the topology
        self.topology[name] = {
                'species': (f'{name}_species',),
                'delta_species': (f'{name}_deltas',),
                'rates': (f'{name}_rates',),
                'globals':(f'volume',),}


    def add_connection(self, source, target, map_function, connector_config = None):
        if source not in self.models:
            raise KeyError(f"source {source} is not the name of a model.")
        if target not in self.models:
            raise KeyError(f"source {target} is not the name of a model.")
        
        source_species = self.models[source].get_species_names()
        target_species = self.models[target].get_species_names()

        if connector_config is None:
            connector_config = {
                'source': source,
                'target': target,
                'source_keys': source_species,
                'target_keys': target_species,
                'map_function': map_function}

        if (source, target) in self.connection_counts:
            count = self.connection_counts[(source, target)]
            count += 1
            self.connection_counts[(source, target)] = count
        else:
            count = 1
            self.connection_counts[(source, target)] = count

        self.connections[f'{source}_{target}_connector_{count}'] = OneWayMap(connector_config)

        #Add process to the topology
        self.topology[f'{source}_{target}_connector_{count}'] = {
            'source_deltas': (f'{source}_deltas',),
            'target_state': (f'{target}_species',),
            'globals': {
                'source_volume': (f'{source}_globals', 'volume'),
                'target_volume': (f'{target}_globals', 'volume'),
            }
        }

def test_connector():
    
    bioscrape_process_1 = Bioscrape(parameters = {
        'sbml_file': 'Notebooks/model1.xml'
        })
    bioscrape_process_3 = Bioscrape(parameters = {
        'sbml_file': 'Notebooks/model3.xml'
        })

    model1_keys = bioscrape_process_1.get_model_species_ids()
    model3_keys = bioscrape_process_3.get_model_species_ids()

    # define the projection
    model1_vector = np.array(['rna_T' == key for key in model1_keys])
    model3_vector = np.array(['rna_T' == key for key in model3_keys])
    projection_1_3 = np.outer(
        model1_vector/np.sum(model1_vector),
        model3_vector/np.sum(model3_vector))
    projection_3_1 = np.outer(
        model3_vector/np.sum(model3_vector),
        model1_vector/np.sum(model1_vector))

        # define map function
    def map_1_3(states):
        input_array = array_from(states['source_deltas'])
        output_array = np.dot(input_array, projection_1_3)
        return array_to(model3_keys, output_array)

    def map_3_1(states):
        input_array = array_from(states['source_deltas'])
        output_array = np.dot(input_array, projection_3_1)
        return array_to(model1_keys, output_array)


    # configuration
    time_step = 1
    models = {
        '1': {
            'time_step': time_step,
            'sbml_file': 'Notebooks/model1.xml'},
        '3': {
            'time_step': time_step,
            'sbml_file': 'Notebooks/model3.xml'},
    }
    connections = [
        {'source': '1', 'target': '3', 'map_function': map_1_3},
        {'source': '3', 'target': '1', 'map_function': map_3_1}
    ]
    # make the composite
    composite = BioscrapeConnector(
        models=models,
        connections=connections,
    )
    

    ## Run a simulation
    # initial state
    config = {
        '1': {
            'rna_T': 10.0
        }
    }
    initial_state = composite.initial_state(config)

    #This occurs after composite.generate() is called
    assert len(composite.models) == 2
    assert len(composite.connections) == 2

    # run a simulation
    sim_settings = {
        'total_time': 10,
        'initial_state': initial_state}
    output = simulate_composer(composite, sim_settings)

    #DNA should be constant
    assert all([output['1_species']['dna_G'][0] == g for g in output['1_species']['dna_G']])

    #RNA should be the same between the two simulations
    assert output['1_species']['rna_T'] == output['3_species']['rna_T']

    #total RNAase should be constant
    RNAase_tot = output['3_species']['protein_RNAase'][0]

    rnas = [output['3_species']['protein_RNAase'][i]+output['3_species']['complex_protein_RNAase_rna_T_'][i] for i in range(len(output['3_species']['complex_protein_RNAase_rna_T_']))]

    #Slight numerical errors are possible, but total RNAase should be nearly constant
    assert  all([np.abs(s-RNAase_tot) < 10**-6 for s in rnas])

def main():
    '''Simulate the composite and plot results.'''
    #make an output directory to save plots
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    output = test_instantiation()

    # plot simulation output
    plot_settings = {}
    plot_simulation_output(output, plot_settings, out_dir)


if __name__ == '__main__':
    main()



"""
run with:
> python -m vivarium_bioscrape.processes.one_way_map
"""

import numpy as np

from vivarium.core.process import Deriver
from vivarium.core.experiment import pp
from vivarium.library.schema import array_from, array_to
from vivarium_bioscrape.processes.bioscrape import Bioscrape
from vivarium_bioscrape.library.mappings import *


class OneWayMap(Deriver):
    name = 'one_way_map'
    defaults = {
        'source': '',
        'target': '',
        'source_keys': [],
        'target_keys': [],
        'map_function': lambda states: {}}

    def __init__(self, parameters=None):
        super(OneWayMap, self).__init__(parameters)
        self.map_function = self.parameters['map_function']

    def initial_state(self):
        return {}

    def ports_schema(self):
        return {
            'source_deltas': {
                species_id: {
                    '_default': 0.0,
                } for species_id in self.parameters['source_keys']},
            'target_state': {
                species_id: {
                    '_default': 0.0,
                } for species_id in self.parameters['target_keys']},
            'globals':{
                f'source_volume': {
                    '_default':1.0
                },
                f'target_volume' : {
                    '_default':1.0
                }
            }}

    def next_update(self, timestep, states):
        output = self.map_function(states)

        #Ensures values never go below 0
        for s in output:
            if states["target_state"][s] + output[s] < 0:
                output[s] = -states["target_state"][s]

        return {
            'target_state': output,
        }

def test_one_way_map():
    bioscrape_process_input = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model3.xml'
        })
    bioscrape_process_output = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model2.xml'
        })

    input_species = bioscrape_process_input.get_model_species() #get_model_species('Notebooks/model3.xml')
    output_species = bioscrape_process_output.get_model_species() #get_model_species('Notebooks/model2.xml')
    input_keys = list(input_species.keys())
    output_keys = list(output_species.keys())

    # define the projection
    input_vector = np.array(['rna' in key for key in input_keys])
    output_vector = np.array(['rna' in key for key in output_keys])
    projection = np.outer(input_vector/np.sum(input_vector), output_vector/np.sum(output_vector))

    # define map function
    def map_function(states):
        input_array = array_from(states['source_deltas'])
        output_array = np.dot(input_array, projection)
        return array_to(output_keys, output_array)

    one_way = OneWayMap({
        'map_function': map_function,
        'source_keys': input_keys,
        'target_keys': output_keys})

    # use input state
    state = {
        'source_deltas': {key: 100 for key in input_species.keys()},
        'target_state': {key: 0.0 for key in output_species.keys()}}

    transform = one_way.next_update(0, state)


def test_one_to_one():
    #Each species is converted into a single other species
    bsp1 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model1.xml'
        })
    bsp2 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model4.xml'
        })

    map_func = one_to_one_map(bsp1, bsp2, {'rna_T':'rna_RNA', 'protein_X':'protein_Protein'})

    f = lambda s: 0 if s != "rna_T" else 1
    r = map_func({'source_deltas':{str(s):f(s) for s in bsp1.get_model_species_ids()}})
    assert r['rna_RNA'] == 1.0
    assert r['protein_Protein'] == 0.0

    r = map_func({'source_deltas':{str(s):f(s)==0 for s in bsp1.get_model_species_ids()}})
    assert r['rna_RNA'] == 0.0
    assert r['protein_Protein'] == 1.0

    r = map_func({'source_deltas':{str(s):1 for s in bsp1.get_model_species_ids()}})
    assert r['rna_RNA'] == 1.0
    assert r['protein_Protein'] == 1.0




def test_one_to_many():
    #One species is converted into many species, evenly.
    #Create Bioscrape Processes
    #Model 1: is a simple transcription translation model
    bsp1 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model1b.xml'
        })

    #Model 4 adds degredation by dilution to proteins and RNAs
    bsp2 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model4.xml'
        })

    #rna_RNA --> rna_T1 + rna_T2
    #protein_Protein --> protein_X1 + protein_X2
    #These species are added evenly
    map_func_21_not_prop = one_to_many_map(bsp2, bsp1,  {"rna_RNA":["rna_T1", "rna_T2"], 
                               "protein_Protein":['protein_X']}, proportional = False)


    f = lambda s: int(s[-1]) if "rna_T" in s else .5
    states =  {
    'source_deltas':{s:10 for s in bsp2.get_model_species_ids()}, 
    'target_state':{s:f(s) for s in bsp1.get_model_species_ids()}}

    r_not_prop = map_func_21_not_prop(states)

    assert r_not_prop['rna_T1'] == 5.
    assert r_not_prop['rna_T2'] == 5.
    assert r_not_prop['protein_X'] == 10
    assert r_not_prop['dna_G1'] == 0

    #These species are added proprotionally to their target concentration
    map_func_21_prop = one_to_many_map(bsp2, bsp1,  {"rna_RNA":["rna_T1", "rna_T2"], 
                               "protein_Protein":['protein_X']}, proportional = True)

    #Degenerate Case: Target concentration is 0
    states0 =  {
    'source_deltas':{s:10 for s in bsp2.get_model_species_ids()}, 
    'target_state':{s:0 for s in bsp1.get_model_species_ids()}}
    r_0 = map_func_21_prop(states0)
    #Should result in the same as the evenly distributed case above
    assert r_0 == r_not_prop

    #Non-generate case
    r_prop = map_func_21_prop(states)
    assert np.allclose(r_prop["rna_T1"], 10./3.)
    assert np.allclose(r_prop["rna_T2"], 2*10./3.)
    assert np.allclose(r_prop["protein_X"], 10.)

def test_many_to_one():
    #Many species are converted into one species
    
    bsp1 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model1b.xml'
        })

    #Model 4 adds degredation by dilution to proteins and RNAs
    bsp2 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model4.xml'
        })

    #Create the map_function using one_to_one_map
    #rna_T1 + rna_T2 --> rna_RNA
    #protein_X1 + protein_X2 --> protein_Protein
    #the many_to_one_map takes the two processes and a dictionary {one_species:[list of many species]}
    map_func_12 = many_to_one_map(bsp1, bsp2, {"rna_RNA":["rna_T1", "rna_T2"], 
                                   "protein_Protein":['protein_X']})

    f = lambda s: int(s[-1]) if "rna_T" in s else .5
    states =  {
    'source_deltas':{s:f(s) for s in bsp1.get_model_species_ids()}, 
    'target_state':{s:0 for s in bsp2.get_model_species_ids()}}

    r = map_func_12(states)

    assert r['rna_RNA'] == 3
    assert r['protein_Protein'] == .5

def test_stochiometrically_compatable():

    #Create Bioscrape Processes
    #Tx Tl Model with multiple ribsome occupancy
    bsp1 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model2.xml'
        })

    #RNA degredation Model
    bsp2 = Bioscrape(parameters = {
        'sbml_file':'Notebooks/model3.xml'
        })

    #Create the map_function using many_to_one_map
    rna_species_1 = [s for s in bsp1.get_model_species_ids() if "rna_T" in s]
    map_func_12 = many_to_one_map(bsp1, bsp2, {"rna_T":rna_species_1})

    #Create the reverse map_function with stochiometric constraints

    #First create a one_to_many parent map function
    parent_map_func_21 = one_to_many_map(bsp2, bsp1,  {"rna_T":rna_species_1})

    #Then create the stochiometric_map

    #Helper function to count the number of ribosomes in a complex
    def ribo_count(s):
        count = 0
        if "x_rna" in s:
            ind = s.index('x_rna')
            count = int(s[ind-1])
        elif "protein_Ribo" in s:
             count = 1
        else:
            count = 0
        return count

    #Ribosomes are created when species with RNA are degraded
    stoch_dict = {s:{'protein_Ribo': -ribo_count(s)} for s in rna_species_1}

    map_func_21 = stochiometric_map(parent_map_func_21, stoch_dict)

    state = {"target_state":{bsp1.get_model_species_ids()[i]:i for i in range(len(bsp1.get_model_species_ids()))}, "source_deltas":{"rna_T":-1, 'protein_RNAase':0, "complex_protein_RNAase_rna_T_":0}}

    r = map_func_21(state)

    #Total amount of RNA degraded should be -1
    assert np.allclose(sum([r[s] for s in rna_species_1]), -1)

    #Total change in ribosome concentration should be 0
    assert np.allclose(sum([r[s]*ribo_count(s) for s in bsp1.get_model_species_ids()]), 0)


if __name__ == '__main__':
    test_one_way_map()
    test_one_to_one()
    test_one_to_many()
    test_many_to_one()
    test_stochiometrically_compatable()



import os
import glob
import setuptools
from distutils.core import setup

with open("README.md", 'r') as readme:
    long_description = readme.read()

setup(
    name='vivarium-bioscrape',
    version='0.0.0.6',
    packages=[
        'vivarium_bioscrape',
        'vivarium_bioscrape.processes',
        'vivarium_bioscrape.composites',
        'vivarium_bioscrape.library'
    ],
    author='William Poole, Eran Agmon, Ryan Spangler',
    author_email='',
    url='https://github.com/vivarium-collective/vivarium-bioscrape',
    license='MIT',
    entry_points={
        'console_scripts': []},
    short_description='Integrates Vivarium with the Bioscrape CRN simulator, allowing Chemical Reaction Network Models to be Vivarium Processes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={},
    include_package_data=True,
    install_requires=[
        'vivarium-core>=0.2.0',
        'bioscrape'
    ],
)
